import shutil

from pathlib import Path

import pandas as pd
import numpy as np
from datetime import datetime
import traceback

import mlinvitrotox.utils.training as ml

from mlinvitrotox.constants import (
    # Paths
    REMOTE_DATA_DIR_PATH,
    INVITRO_INPUT_DIR_PATH,
    CONFIG_DIR_PATH,
    FPS_OUTPUT_DIR_PATH,
    VALIDATION_FPS_OUTPUT_DIR_PATH,
    MASSBANK_OUTPUT_DIR_PATH,
    MODELS_RESULTS_DIR_PATH,  # 31
    # Folders
    REMOTE_PYTCPL_FOLDER,  # 7
    PYTCPL_INPUT_FOLDER,  # 7
    # Files
    SELECTED_AEIDS,  # 30
    SELECTED_CHEMICALS,  # part of 31
    PYTCPL_OVERVIEW_FILE,  # 8
    ICE_REFERENCES,  # 9
    CONFIG_FILE,  # C
    CONFIG_MODELS_FILE,  # C
    TRUE_FPS_TRAINING,  # 13
    VALIDATION_PRED_FPS_DTXSID,  # 20
    VALIDATION_SELECTED_ABSINDEX,  # 21
)

# !! in config_classiffication, use complete hyperparameter grid again
# %%

# Get pytcpl output from remote repository
remote_pytcpl_folder = Path(REMOTE_DATA_DIR_PATH) / REMOTE_PYTCPL_FOLDER
pytcpl_input_folder = Path(INVITRO_INPUT_DIR_PATH) / PYTCPL_INPUT_FOLDER
if 0:
    ml.empty_pytcpl_input_directory(pytcpl_input_folder)
    ml.copy_pytcpl_parquet_files(remote_pytcpl_folder, pytcpl_input_folder)

# %%

# Set main folder to log results
logs_folder = Path(MODELS_RESULTS_DIR_PATH)

# Get ML configuration
config_path = Path(CONFIG_DIR_PATH) / CONFIG_FILE
config_models_path = Path(CONFIG_DIR_PATH) / CONFIG_MODELS_FILE
CONFIG, _, LOGGER, RUN_FOLDER, LOG_PATH = ml.load_config(config_path, logs_folder)

# set some configuration values
target_variable = CONFIG["target_variable"]
TARGET_VARIABLE = ml.init_target_variable(target_variable)
ml_algorithm = "classification"
CONFIG_ESTIMATORS = ml.init_ml_algo(ml_algorithm, str(config_models_path))

# load ICE references file and pytcpl overview file
ice_input_file = Path(INVITRO_INPUT_DIR_PATH) / ICE_REFERENCES
ice_df = pd.read_csv(ice_input_file)
# TODO copy from remote pytcpl repo
pytcpl_overview_file = Path(INVITRO_INPUT_DIR_PATH) / PYTCPL_OVERVIEW_FILE
pytcpl_df = pd.read_csv(pytcpl_overview_file)
assay_info_df = ml.merge_assay_info(pytcpl_df, ice_df)

# load aeid files
aeid_info_df = ml.get_aeid_files(
    pytcpl_input_folder,
    target_variable,
    assay_info_df,
)

# Select aeids
available_aeid_list = aeid_info_df["aeid"].tolist()
available_aeid_list = [str(aeid) for aeid in sorted(available_aeid_list)]
aeid_list = ml.select_aeids_for_training(
    pytcpl_input_folder,
    aeid_info_df,
    aeids=available_aeid_list,
    #    mech_target="Androgen"    # A mechanistic target can be specified.
)
LOGGER.info(f"AEIDs selected for training: {aeid_list}")
LOGGER.info(f"Number of target aeids: {len(aeid_list)}")

# Get true fingerprint data (from training)
LOGGER.info("get_fingerprint_df function is running")
LOGGER.info("#" * 60)
fps_input_file = Path(FPS_OUTPUT_DIR_PATH) / TRUE_FPS_TRAINING
fps_df = ml.get_fingerprint_df(fps_input_file)

# Iterate through aeids_target_assays and launch each iteration in a separate process
selected_aeid_list = []
# for aeid in aeid_list:
for aeid in aeid_list[:2]:
    # log mechanistic target and aenm
    mech_target = aeid_info_df[aeid_info_df["aeid"].astype(str) == aeid][
        "MechanisticTarget"
    ].iloc[0]
    acen = aeid_info_df[aeid_info_df["aeid"].astype(str) == aeid][
        "assay_component_endpoint_name"
    ].iloc[0]
    LOGGER.info(
        f"Included AEID: {aeid}, MechanisticTarget: {mech_target}, assay component endpoint name: {acen}"
    )

    try:
        # init aeid
        ml.init_aeid(aeid)
        LOGGER.info(
            f"Start ML pipeline for assay ID: {aeid} and target variable: {target_variable}"
        )
        LOGGER.info("#" * 60)
        # get toxicity data
        LOGGER.info("process_single_aeid function is running")
        LOGGER.info("#" * 60)
        ice_omit_filtering = CONFIG["apply"]["ice_omit_filtering"]
        activity_threshold = CONFIG["activity_threshold"]
        assay_df = ml.process_single_aeid(
            pytcpl_input_folder,
            aeid,
            target_variable,
            ice_omit_filtering,
            activity_threshold,
        )
        LOGGER.info(f"Activity threshold: >= {activity_threshold} is active \n")

        # merge chemical ids in both dataframes
        LOGGER.info("merge_assay_and_fingerprint_df function is running")
        LOGGER.info("#" * 60)
        selected_chemicals_file = SELECTED_CHEMICALS
        merged_df = ml.merge_assay_and_fingerprint_df(
            assay_df,
            fps_df,
            selected_chemicals_file,
        )
        merged_df[target_variable] = merged_df[target_variable].astype(int)
        # LOGGER.info(f" Columns in the df: {merged_df.columns}")
        LOGGER.info(f"Shape of the merged df: {merged_df.shape}")
        LOGGER.info(f"hitcall column type: {merged_df[target_variable].dtype}")

        # filter aeids with little data and few positive labels
        dtxsid_count = 100
        npos_count = 10
        if len(merged_df) < dtxsid_count:
            LOGGER.info(f"Less than {dtxsid_count} chemicals for aeid: {aeid}")
            LOGGER.info(f"Not used for modeling: {aeid}")
            continue
        if np.sum(merged_df[target_variable] == 1) < npos_count:
            LOGGER.info(f"Less than {npos_count} positive labels for aeid: {aeid}")
            LOGGER.info(f"Not used for modeling: {aeid}")
            continue
        LOGGER.info(f"Used for modeling: {aeid}")
        selected_aeid_list.append(aeid)

        # cluster molecules within hitcall groups
        LOGGER.info("clustering is running")
        LOGGER.info("#" * 60)
        n_cls_0 = CONFIG["apply"]["n_cls_0"]
        n_cls_1 = CONFIG["apply"]["n_cls_1"]
        random_state = CONFIG["random_state"]
        clust_df = ml.clustering(
            merged_df,
            target_variable,
            n_cls_0,
            n_cls_1,
            random_state,
        )
        LOGGER.info(f"Shape of the clustered dataframe: {clust_df.shape}")
        LOGGER.info(f"Counts of 0s and 1s: {clust_df[target_variable].value_counts()}")

        # filter for indexes with good validation performance
        sirius_quality_filtering = CONFIG["apply"]["sirius_quality_filtering"]
        if sirius_quality_filtering == 1:
            LOGGER.info("filter_with_sirius_quality_indexes function is running")
            LOGGER.info("#" * 60)
            indexes_path = (
                Path(VALIDATION_FPS_OUTPUT_DIR_PATH) / VALIDATION_SELECTED_ABSINDEX
            )
            filtered_df = ml.filter_with_sirius_quality_indexes(
                clust_df,
                indexes_path,
                target_variable,
            )
        else:
            filtered_df = clust_df.copy()
        LOGGER.info(f"Filtered dataframe shape: {filtered_df.shape}")

        # split data into training and validation data
        LOGGER.info("split_training_validation_data function is running")
        LOGGER.info("#" * 60)
        pred_fps_file = Path(MASSBANK_OUTPUT_DIR_PATH) / VALIDATION_PRED_FPS_DTXSID
        pred_fps_df = pd.read_csv(pred_fps_file)
        LOGGER.info(f"Data frame shape of pred_fps_dtxsid: {pred_fps_df.shape}")
        # LOGGER.info(f"Data frame columns in pred_fps_dtxsid: {pred_fps_df.columns}")
        (
            df_train_test,
            df_validation_true,
            df_validation_pred,
        ) = ml.split_training_validation_data(
            filtered_df,
            pred_fps_df,
            target_variable,
        )

        # Partition data into X and y and respective massbank validation set
        # (massbank validation set evaluates generalization to unseen spectral data)
        LOGGER.info("partition training and validation data is running")
        LOGGER.info("#" * 60)
        (X, y, cluster, feature_names) = ml.partition_training_data(
            df_train_test,
            target_variable,
        )

        (
            X_mbval_structure,
            X_mbval_sirius,
            y_mbval,
        ) = ml.partition_validation_data(
            df_validation_true, df_validation_pred, target_variable, feature_names
        )
        LOGGER.info(f"Training X shape: {X.shape}")
        LOGGER.info(f"X_mbval_structure shape: {X_mbval_structure.shape}")
        LOGGER.info(f"X_mbval_sirius: {X_mbval_sirius.shape}")
        if not X.columns.equals(X_mbval_structure.columns) or not X.columns.equals(
            X_mbval_sirius.columns
        ):
            LOGGER.warning(
                "Columns in X, X_mbval_structure, and X_mbval_sirius are not the same!"
            )

        # Split data into train test set
        # (Stratified split by clusters)
        # (Single member clusters are in the training set)
        LOGGER.info("split_training_data is running")
        LOGGER.info("#" * 60)
        train_test_split_ratio = CONFIG["train_test_split_ratio"]
        X_train, y_train, X_test, y_test = ml.split_training_data(
            X,
            y,
            cluster,
            train_test_split_ratio,
            random_state,
        )

        # Apply oversampling
        oversampling = CONFIG["apply"]["oversampling"]
        if oversampling == 1:
            LOGGER.info(" oversampling is running")
            LOGGER.info("#" * 60)
            X_train_os, y_train_os = ml.handle_oversampling(
                X_train,
                y_train,
                random_state,
            )
        else:
            X_train_os = X_train.copy()
            y_train_os = y_train.copy()

        # Prepare feature selection
        feature_selection = CONFIG["apply"]["feature_selection"]
        remove_lowvariance_features = CONFIG["apply"]["remove_lowvariance_features"]
        remove_highcorrelation_features = CONFIG["apply"][
            "remove_highcorrelation_features"
        ]
        variance_threshold = CONFIG["feature_selection"]["variance_threshold"]
        correlation_threshold = CONFIG["feature_selection"]["correlation_threshold"]
        preprocessing_pipelines = ml.build_preprocessing_pipeline(
            feature_selection,
            remove_lowvariance_features,
            remove_highcorrelation_features,
            variance_threshold,
            correlation_threshold,
            random_state,
        )

        for preprocessing_pipeline in preprocessing_pipelines:
            ml.init_preprocessing_pipeline(preprocessing_pipeline)

            (
                feature_names,
                X_train_os,
                X_train,
                X_test,
                X_mbval_structure,
                X_mbval_sirius,
            ) = ml.preprocess_all_sets(
                preprocessing_pipeline,
                feature_names,
                X_train_os,
                y_train_os,
                X_train,
                X_test,
                X_mbval_structure,
                X_mbval_sirius,
            )

            # Get the label counts
            ml.log_binarized_label_count(y, "TOTAL", activity_threshold)
            ml.log_binarized_label_count(y_train_os, "TRAIN", activity_threshold)
            ml.log_binarized_label_count(y_test, "TEST", activity_threshold)
            ml.log_binarized_label_count(
                y_mbval, "MassBank VALIDATION", activity_threshold
            )

            LOGGER.info("Run pipeline for all estimators:\n")
            # Build for each estimator a pipeline according to the configurations in the config file
            for estimator in CONFIG_ESTIMATORS["estimators"]:
                start_time = datetime.now()

                # Init a new folder for this estimator
                estimator_name = estimator["name"]
                ESTIMATOR_PIPELINE = ml.init_estimator_pipeline(estimator_name)
                LOGGER.info(f"Apply {ESTIMATOR_PIPELINE}")

                ## Training
                # Build the pipeline for the current estimator with the specified parameter grid
                pipeline = ml.build_pipeline(estimator, random_state)

                # Perform grid search cross-validation
                # (Note: CV on TRAINING set with RepeatedStratifiedKFold)
                LOGGER.info("Start Grid Search Cross-Validation..")
                n_splits = CONFIG["grid_search_cv"]["n_splits"]
                n_repeats = CONFIG["grid_search_cv"]["n_repeats"]
                grid_search = ml.perform_grid_search_cv(
                    X_train_os,
                    y_train_os,
                    estimator,
                    pipeline,
                    n_splits,
                    n_repeats,
                    random_state,
                )
                LOGGER.info("Training Done.\n")

                # Save best estimator (estimator with best performing parameters from grid search)
                best_estimator = grid_search.best_estimator_
                ml.save_model(best_estimator, "train")

                ## Validation
                # Predict on the test set with the best estimator (X_test, y_test is unseen)
                LOGGER.info("Start Internal Testing")
                validation_set = ml.init_validation_set("test")
                LOGGER.info("+" * 60)
                LOGGER.info(f"Predict ({validation_set})")
                default_threshold = CONFIG["threshold_moving"]["default_threshold"]
                threshold_moving = CONFIG["apply"]["threshold_moving"]
                target_tpr = CONFIG["threshold_moving"]["target_tpr"]
                target_tnr = CONFIG["threshold_moving"]["target_tnr"]
                cost_tpr = CONFIG["threshold_moving"]["cost_tpr"]
                cost_fpr = CONFIG["threshold_moving"]["cost_fpr"]
                ml.predict_and_report_classification(
                    X_test,
                    y_test,
                    best_estimator,
                    default_threshold,
                    threshold_moving,
                    target_tpr,
                    target_tnr,
                    cost_tpr,
                    cost_fpr,
                )
                LOGGER.info("Internal Testing Done.\n")

                # Retrain the best estimator from GridSearchCV with train and test data
                ml.init_estimator_pipeline(estimator_name)
                LOGGER.info("Retrain on train/test set..\n")
                X_combined = np.vstack((X_train, X_test))
                y_combined = np.concatenate((y_train, y_test))
                if oversampling == 1:
                    LOGGER.info(" oversampling of train/test dataset is running")
                    LOGGER.info("#" * 60)
                    X_combined_os, y_combined_os = ml.handle_oversampling(
                        X_combined,
                        y_combined,
                        random_state,
                    )
                else:
                    X_combined_os = X_combined.copy()
                    y_combined_os = y_combined.copy()
                best_estimator.fit(
                    X_combined_os,
                    y_combined_os,
                )
                ml.save_model(best_estimator, "train_test")

                # Predict on the 1. "true Massbank" and 2. "SIRIUS predicted" validation set
                LOGGER.info("Start MassBank Validation")
                validation_set_names = ["structure", "sirius"]
                validation_sets = [
                    X_mbval_structure,
                    X_mbval_sirius,
                ]
                for name, X_mbval in zip(validation_set_names, validation_sets):
                    validation_set = ml.init_validation_set(f"mb_val_{name}")
                    LOGGER.info("+" * 60)
                    LOGGER.info(f"Predict ({validation_set})")
                    ml.predict_and_report_classification(
                        X_mbval,
                        y_mbval,
                        best_estimator,
                        default_threshold,
                        threshold_moving,
                        target_tpr,
                        target_tnr,
                        cost_tpr,
                        cost_fpr,
                    )

                LOGGER.info("MassBank Validation Done.\n")

                # Retrain the estimator on full data for future predictions
                ml.init_estimator_pipeline(estimator_name)
                LOGGER.info("Retrain on full data..\n")
                X_all = np.vstack((X_combined, X_mbval_structure))
                y_all = np.concatenate((y_combined, y_mbval))
                if oversampling == 1:
                    LOGGER.info(" oversampling of full dataset is running")
                    LOGGER.info("#" * 60)
                    X_all_os, y_all_os = ml.handle_oversampling(
                        X_all,
                        y_all,
                        random_state,
                    )
                else:
                    X_all_os = X_all.copy()
                    y_all_os = y_all.copy()
                best_estimator.fit(X_all_os, y_all_os)
                ml.save_model(best_estimator, "full_data")

                # Get feature importances
                # (only implemented for XGB and RandomForest)
                ml.get_feature_importance_if_applicable(best_estimator, feature_names)

                # Time elapsed
                elapsed = round((datetime.now() - start_time).total_seconds(), 2)
                LOGGER.info(f"Done {estimator['name']} >> {elapsed} seconds.\n\n\n")

    except Exception as e:
        traceback_info = traceback.format_exc()
        ml.report_exception(e, traceback_info, aeid)

        # Write a failed flag file to the aeid folder
        ml.init_aeid(aeid)  # re-init aeid's DUMP_FOLDER
        ml.add_status_file("failed")

# Calculate the total elapsed time
LOGGER.info(f"Finished all: Total time >> { ml.get_total_elapsed_time()}\n")

# store aeid file
aeid_df = pd.DataFrame(selected_aeid_list, columns=["aeid"]).astype("int")
aeid_df = pd.merge(aeid_df, aeid_info_df, on="aeid", how="left")
selected_aeid_file = Path(LOG_PATH) / SELECTED_AEIDS
aeid_df.sort_values("aeid").round(5).to_csv(selected_aeid_file, index=False)
print(f"Processed and saved to file: {selected_aeid_file}")

# copy true training fps file to run folder
shutil.copy2(fps_input_file, LOG_PATH)
print(f"Copied true training fps file to {LOG_PATH}")



import click

from .apply.extractsiriusdata import extract_sirius_data
from .apply.loaduserdata import load_user_data
from .apply.runmodels import run_models

# from .apply.startapp import start_app
from .train.evaluatemodels import run_evaluation
from .train.providemodels import export_model


@click.group()
def main():
    """main"""
    pass


main.add_command(extract_sirius_data)
main.add_command(load_user_data)
main.add_command(run_models)
# main.add_command(start_app)


@main.group("train")
def train():
    """train"""
    pass


train.add_command(run_evaluation)
train.add_command(export_model)

if __name__ == "__main__":
    main()



# Data processing: 3) extract data from sirius

import os

from pathlib import Path

import pandas as pd

from tqdm import tqdm

from mlinvitrotox.constants import (
    # IDs
    MASSBANK_ID,
    SIRIUS_ID,
)


def check_sirius_version(
    df_csi_fpt,
):
    """
    This is a simple check to ensure that the SIRIUS output
    used to train the mlinvitrotox models
    is comparable to the SIRIUS output of the user.

    """

    # can be either provided directly as a file or as a path
    if isinstance(df_csi_fpt, (str, Path)):
        df_csi_fpt = pd.read_csv(df_csi_fpt, sep="\t")

    if not len(df_csi_fpt) == 3878:
        print(
            "Error: The HRMS data have been processed with a different SIRIUS version than used in MLinvitroTox."
        )
        return False
    else:
        return True


def _format_column_name(file_path):
    parts = file_path.split(os.sep)
    main_part_parts = parts[-2].split("_", 1)
    main_part = main_part_parts[0] + "_" + main_part_parts[1].replace("_", "-")
    last_part = parts[-1].split(".")[0]
    return f"{main_part}_{last_part}"


def compute_fps_data(file_path, absoluteIndex, threshold):
    # load fingerprint
    col_name = _format_column_name(str(file_path))
    df = pd.read_csv(file_path, header=None, names=[col_name])

    # check if it's a fingerprint from the positive mode
    if len(df) != 3878:
        print("The fingerprint does not contain exactly 3878 bits", col_name)
        print("It is not from the positive mode and will not be processed.")
        return {}

    # set the absoluteIndexes from df_csi
    df.index = absoluteIndex

    # apply threshold to generate binary features
    df[col_name] = df[col_name].apply(lambda x: 0 if x < threshold else 1)

    # Check corresponding .info file for 'TopCSIScore'
    dict_info = {}
    info_file_path = str(file_path).replace(".fpt", ".info")
    if os.path.exists(info_file_path):
        with open(info_file_path, "r") as info_file:
            lines = info_file.readlines()
            if lines and any(["TopCSIScore" in line for line in lines]):
                dict_info["TopCSIScore"] = "available"
            else:
                dict_info["TopCSIScore"] = "none"

    # get formula rank
    formula_file_path = (
        "/".join(str(file_path).split("/")[:-1]) + "/formula_candidates.tsv"
    )
    if os.path.exists(formula_file_path):
        formula_df = pd.read_csv(formula_file_path, sep="\t")
        precursor = col_name.split("_")[2]
        adduct = col_name.split("_")[3]
        formularank = formula_df[
            (formula_df["precursorFormula"] == precursor)
            & (formula_df["adduct"].str.replace(" ", "") == adduct)
        ]["formulaRank"].iloc[0]
        dict_info["formulaRank"] = formularank

    # concatenate fingerprint with info
    df_info = pd.Series(dict_info).to_frame(name=col_name)
    df_concat = pd.concat((df, df_info), axis=0)
    return {col_name: df_concat}


def read_fps_files(
    input_fpt_directory,
    df_csi,
    threshold=0.5,
    max_workers=None,
):
    # can be either provided directly as a file or as a path
    if isinstance(df_csi, (str, Path)):
        df_csi = pd.read_csv(df_csi, sep="\t")

    # initialize
    fps_data = {}

    # get file paths for .fpt files
    all_fps_files = [
        Path(root) / f
        for root, dirs, files in os.walk(input_fpt_directory)
        for f in files
        if f.endswith(".fpt")
    ]
    print(f"Found {len(all_fps_files)} .fpt files.")

    absoluteIndex = df_csi["absoluteIndex"].values

    # process files
    if max_workers is None or max_workers > 1:
        import concurrent.futures

        with concurrent.futures.ProcessPoolExecutor(
            max_workers=max_workers
        ) as executor:
            future_to_file_path = [
                executor.submit(compute_fps_data, file_path, absoluteIndex, threshold)
                for file_path in all_fps_files
            ]
            # future_to_file_path = {executor.submit(compute_fps_data, file_path, absoluteIndex, threshold): file_path for file_path in all_fps_files}
            for future in tqdm(
                concurrent.futures.as_completed(future_to_file_path),
                total=len(all_fps_files),
            ):
                # file_path = future_to_file_path[future]
                fps_column = future.result()
                if fps_column:
                    # merge two dictonaries
                    fps_data |= fps_column
    else:
        for file_path in tqdm(all_fps_files, desc="Processing files"):
            fps_column = compute_fps_data(file_path, absoluteIndex, threshold)
            if fps_column:
                # merge two dictonaries
                fps_data |= fps_column

    # store
    if fps_data:
        fps_df = pd.concat(fps_data.values(), axis=1)
        fps_df.columns = list(fps_data.keys())
        print(f"Dataframe with imported fingerprints: {fps_df.shape[1]} entries")

        # transpose and only keep formulaRank 1 entries
        fps_df_transposed = fps_df.T
        fps_df_transposed.columns = [
            str(col).zfill(4) for col in fps_df_transposed.columns
        ]
        fps_df_transposed.reset_index(inplace=True)
        fps_df_transposed.rename(columns={"index": "features"}, inplace=True)
        if "formulaRank" in fps_df_transposed.columns:
            fps_df_transposed = fps_df_transposed[
                fps_df_transposed["formulaRank"] == 1
            ].copy()

        # extract information from features column
        fps_df_transposed = _extract_info_from_features_column(
            fps_df_transposed,
        )

        # sort to have reproducible output
        fps_df_sorted = fps_df_transposed.sort_values(["features"])

        return fps_df_sorted

    else:
        print("No valid data extracted from .fpt files.")
        return pd.DataFrame()


def keep_selected_fps_bits(
    df_predfps,
    df_truefps,
):
    # only keep fingerprint bits used in training data
    df_truefps.columns = df_truefps.columns.astype(str)
    df_predfps.columns = df_predfps.columns.astype(str)
    columns_to_keep = [
        col
        for col in df_predfps.columns
        if col
        in [
            "features",
            "TopCSIScore",
            "formulaRank",
            SIRIUS_ID,
            MASSBANK_ID,
            "sirius_run",
            "formula",
            "adduct",
        ]
    ]
    columns_to_keep += df_predfps.columns[
        df_predfps.columns.isin(df_truefps.columns)
    ].tolist()
    df_predfps_selected = df_predfps.loc[:, columns_to_keep]

    print(
        f"Dataframe with predicted fingerprints: {df_predfps_selected.shape[0]} entries, {df_predfps_selected.shape[1]} fingerprint bits"
    )

    return df_predfps_selected


def _extract_info_from_features_column(df):
    # Split features
    split_df = df["features"].str.split("_", expand=True)

    # Assign columns based on the number of parts found in the split
    df[SIRIUS_ID] = split_df[0]
    df["sirius_run"] = split_df[1]
    df["formula"] = split_df[2]

    # For rows with more than three splits, assign the fourth part to 'adduct', else None
    if len(split_df.columns) <= 3:
        print(split_df)
    df["adduct"] = split_df[3] if len(split_df.columns) > 3 else None

    return df


def filter_mbval_data_by_accession(
    df,
    input_massbank_validation,
):
    # add adjusted sirius run column
    df["adjusted_sirius_run"] = (
        df["sirius_run"].str.split("-", n=3).str[:3].str.join("-")
    )

    # load MassBank validation file
    df_massbank_validation = pd.read_csv(input_massbank_validation)
    df_cut = df_massbank_validation[["ACCESSION", MASSBANK_ID]]

    # merge with input df
    merged_df = pd.merge(
        df,
        df_cut,
        left_on="adjusted_sirius_run",
        right_on="ACCESSION",
        how="inner",
    )
    return merged_df


# aggregate fingerprints for MassBank validation data
def custom_agg(x):
    if x.name.isdigit():
        return max(x)
    else:
        return list(x.unique())



import sys

from pathlib import Path

import mlinvitrotox.utils.get_predicted_fps as get_predfps
import mlinvitrotox.utils.predict as predict

from mlinvitrotox.constants import (
    # Files
    SIRIUS_FPS_DEFINITIONS,
)

import click


@click.command("load")
@click.option(
    "--sirius_input_folder",
    "-i",
    help="path to SIRIUS files",
    type=click.Path(exists=True, path_type=Path),
)
@click.option(
    "--output_file",
    "-o",
    default=None,
    help="path to csv file where the processed output should be stored. It will be created if it does not exist.",
)
@click.option(
    "--max_workers",
    "-w",
    default=None,
    type=int,
    help="Number of workers/process to use for loading the data",
)
def load_user_data(sirius_input_folder, output_file, max_workers=None):
    # input path for SIRIUS files
    print("SIRIUS input folder with .fpt files:", sirius_input_folder)

    # check SIRIUS version
    matching_sirius_version = get_predfps.check_sirius_version(
        sirius_input_folder / SIRIUS_FPS_DEFINITIONS,
    )
    if not matching_sirius_version:
        "Error: This data cannot be processed."
        sys.exit()

    # read SIRIUS .fpt files
    df_fps = get_predfps.read_fps_files(
        sirius_input_folder,
        sirius_input_folder / SIRIUS_FPS_DEFINITIONS,
        threshold=0.5,
        max_workers=max_workers,
    )

    # sort and reset index
    df_fps = df_fps.sort_index().reset_index(drop=True)
    print(f"Final {sirius_input_folder.name} dataframe: {df_fps.shape[0]} entries")

    # set output file path
    if output_file is None:
        output_file = (
            Path("./results") / sirius_input_folder.name / "sirius-pred-fps.csv"
        )
    output_file = Path(output_file)
    output_file.parent.mkdir(parents=True, exist_ok=True)

    # store output
    print(f"Writing to file {output_file}")
    if "features" in df_fps.columns:
        df_fps = predict.sort_by_column(df_fps, "features")
    df_fps.to_csv(output_file, index=False)


if __name__ == "__main__":
    load_user_data()



import os

from pathlib import Path

import pandas as pd

import streamlit as st

from mlinvitrotox.constants import (
    # MODELS_RESULTS_DIR_PATH,
    # Files
    SELECTED_AEIDS,  # 30
)

# Add a dropdon to the sidebar: Select run folder
logs_folder = Path("models")
# logs_folder = Path(MODELS_RESULTS_DIR_PATH)
run_folder = st.sidebar.selectbox(
    "Select Run", [run_folder for run_folder in os.listdir(logs_folder)][::-1]
)
run_folder_path = logs_folder / run_folder

# Load logged assay information overview
assay_info_path = run_folder_path / SELECTED_AEIDS
assay_info_df = pd.read_csv(assay_info_path)
aeid_list = sorted(assay_info_df["aeid"].to_list())

# Add a dropdown to the sidebar: Select target variable
target_variable_list = os.listdir(run_folder_path)
target_variable_list = [
    d
    for d in target_variable_list
    if os.path.isdir(run_folder_path / d) and "hitcall" in d
]
if len(target_variable_list) > 1:
    target_variable = st.sidebar.selectbox(
        "Select Target Variable", target_variable_list
    )
else:
    target_variable = target_variable_list[0]
    st.sidebar.write(f"Target variable: {target_variable}")

# Add a dropdown to the sidebar: Select algorithm
target_variable_path = run_folder_path / target_variable
ml_algorithm_list = os.listdir(target_variable_path)
ml_algorithm_list = [
    d for d in ml_algorithm_list if os.path.isdir(target_variable_path / d)
]
if len(ml_algorithm_list) > 1:
    ml_algorithm = st.sidebar.selectbox(
        "Select ML Algorithm",
        ml_algorithm_list,
    )
else:
    ml_algorithm = ml_algorithm_list[0]
    st.sidebar.write(f"Algorithm: {ml_algorithm}")

# Add a dropdown to the sidebar: Select aeid
aeid = str(
    st.sidebar.selectbox(
        "Select AEID",
        aeid_list,
    )
)

# Add a dropdown to the sidebar: Select preprocessing model
aeid_path = target_variable_path / ml_algorithm / aeid
preprocessor_list = os.listdir(aeid_path)
preprocessor_list = [d for d in preprocessor_list if os.path.isdir(aeid_path / d)]
results_available = True
if len(preprocessor_list) == 0:
    st.write("No results for this aeid")
    results_available = False
elif len(preprocessor_list) > 1:
    preprocessor = st.sidebar.selectbox(
        "Select feature selection model", preprocessor_list
    )
else:
    preprocessor = preprocessor_list[0]
    st.sidebar.write(f"Feature selection: {preprocessor.split('_')[2]}")

if results_available:
    # Add a dropdown to the sidebar: Select estimator model
    preprocessor_path = aeid_path / preprocessor
    estimator_list = os.listdir(preprocessor_path)
    estimator_list = [d for d in estimator_list if os.path.isdir(preprocessor_path / d)]
    if len(estimator_list) > 1:
        estimator = st.sidebar.selectbox("Select Estimator Model", estimator_list)
    else:
        estimator = estimator_list[0]
        st.sidebar.write(f"Model: {estimator}")

    # Add a dropdown to the sidebar: Select validation type
    # TODO unify names with validation summary
    estimator_path = preprocessor_path / estimator
    val_type_list = os.listdir(estimator_path)
    val_type_list = [d for d in val_type_list if os.path.isdir(estimator_path / d)]
    val_type = st.sidebar.selectbox(
        "Select Validation Type",
        val_type_list[::-1],
    )

    st.title("Validation Results (using 4 different classification thresholds)")
    info_data = {
        "Target Variable": [target_variable],
        "ML Algorithm": [ml_algorithm],
        "aeid": [aeid],
        "Feature Selection": [preprocessor.split("_")[2]],
        "Estimator": [estimator],
        "Validation Set": [val_type],
    }
    info_df = pd.DataFrame(info_data)
    st.dataframe(info_df, hide_index=True, use_container_width=True)

    # Plot confusion matrices for all 4 classification thresholds
    val_type_path = estimator_path / val_type
    confusion_matrices_paths = {}
    roc_curve_paths = {}
    reports = {}
    threshold_names = []
    threshold_values = []
    # TODO do not hardcode the classification threshold entries
    for classification_threshold in ["default", "optimal", "tnr", "tpr"]:
        cm_path = val_type_path / f"cm_{classification_threshold}.svg"
        confusion_matrices_paths[classification_threshold] = cm_path

        report_path = val_type_path / f"report_{classification_threshold}.csv"
        reports[classification_threshold] = (
            pd.read_csv(report_path)
            .reset_index(drop=True)
            .rename(columns={"Unnamed: 0": "class"})
        )
        threshold_names.append(classification_threshold)

    roc_path = val_type_path / "roc_curve.svg"
    pr_path = val_type_path / "pr_curve.svg"

    with st.container():
        col1, col2 = st.columns(2)
        with col1:
            st.header("ROC curve")
            st.image(open(roc_path).read())
        with col2:
            st.header("Precision recall curve")
            st.image(open(pr_path).read())

    st.divider()

    # Create a 2x2 grid using columns
    st.header("Confusion matrices for different thresholds")
    with st.container():
        col1, col2 = st.columns(2)
        with col1:
            i = 0
            st.subheader("Default = 0.5")
            threshold = threshold_names[i]
            cm_path = confusion_matrices_paths[threshold]
            st.image(open(cm_path).read())
            with st.expander("Show classification report"):
                st.dataframe(reports[threshold])
        with col2:
            i = 1
            st.subheader("TNR ≈ 0.5")
            threshold = threshold_names[i]
            cm_path = confusion_matrices_paths[threshold]
            st.image(open(cm_path).read())
            with st.expander("Show classification report"):
                st.dataframe(reports[threshold])

    st.divider()

    with st.container():
        col1, col2 = st.columns(2)
        with col1:
            i = 3
            st.subheader("Cost(TPR, TNR) = 2*(1-TPR) + (1-TNR)")
            threshold = threshold_names[i]
            cm_path = confusion_matrices_paths[threshold]
            st.image(open(cm_path).read())
            with st.expander("Show classification report"):
                st.dataframe(reports[threshold])
        with col2:
            i = 2
            st.subheader("TPR ≈ 0.5")
            threshold = threshold_names[i]
            cm_path = confusion_matrices_paths[threshold]
            st.image(open(cm_path).read())
            with st.expander("Show classification report"):
                st.dataframe(reports[threshold])

    st.divider()



import glob
import zipfile

from pathlib import Path

import filetype
from tqdm import tqdm
import click


def extract_and_find_zips(zip_path, output_dir=None):
    # Convert paths to Path objects
    zip_path = Path(zip_path)

    if not zip_path.is_file() and output_dir is not None:
        raise RuntimeError("If SIRIUS input is a folder, output dir must be omitted")

    if output_dir is None:
        output_dir = zip_path.with_suffix("")

    output_dir = Path(output_dir)
    # Ensure the output directory exists
    output_dir.mkdir(parents=True, exist_ok=True)

    # Extract the main zip file
    if zip_path.is_file():
        print(f"Unzipping {zip_path}")
        with zipfile.ZipFile(zip_path, "r") as zip_ref:
            zip_ref.extractall(output_dir)

    # Recursively find and extract other zip files using glob
    def extract_nested_zips(directory):
        # Use glob to recursively find all files in the directory
        file_paths = glob.glob(f"{directory}/**/*", recursive=True)
        for file_path in tqdm(file_paths, desc="Processing files"):
            file = Path(file_path)
            if file.is_file():
                # Check if the file is a zip file
                kind = filetype.guess(file)
                if kind and kind.extension == "zip":
                    # Extract nested zip file
                    with zipfile.ZipFile(file, "r") as nested_zip_ref:
                        nested_zip_ref.extractall(file.parent)

    # Start the recursive search and extraction process
    extract_nested_zips(output_dir)

    print(f"Extracted the SIRIUS data {zip_path} to {output_dir}/")


@click.command("extract")
@click.option(
    "--sirius_input",
    "-i",
    help="path to SIRIUS ZIP file or folder of SIRIUS content",
    type=click.Path(exists=True, path_type=Path),
)
@click.option(
    "--output_folder",
    "-o",
    default=None,
    help="path to directory where the processed output should be stored. It will be created if it does not exist.",
)
def extract_sirius_data(sirius_input, output_folder):
    """
    Extract data from SIRIUS

    """
    extract_and_find_zips(sirius_input, output_folder)


if __name__ == "__main__":
    extract_sirius_data()
